/*****
*
* Copyright (C) 2001 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Written by Yoann Vandoorselaere <yoann@mandrakesoft.com>
*
*****/


#include <stdio.h>
#include <stdlib.h>

#include "protocol.h"
#include "rpc-prv.h"


int rpc_plugin_id;
port_list_t port_list;


static int add_rpc_rule(rpc_test_t *ptr, rules_t **rules) 
{
        int ret;
        rule_t *rule;

        rule = make_new_rule(0, NULL);
        if ( ! rule )
                return -1;
        
        ret = add_rule_leaf_match(rule, 1, ptr, &match_rpc_rule);
        if ( ret < 0 )
                return -1;
        
        *rules = make_new_rules(rule, NULL);
        if ( ! *rules )
                return -1;
        
        return 0;
}



static int parse_rpc_rule(parameters_t *plist, rules_t **rules)
{
        int i;
        rpc_test_t *ptr;
        
        ptr = calloc(1, sizeof(rpc_test_t));
        if ( ! ptr ) {
                log(LOG_ERR, "memory exhausted.\n");
                return -1;
        }

        for ( i = 0; plist != NULL; plist = plist->next ) {

                /*
                 * This parameter is empty.
                 */
                if ( *plist->str == '*' ) {
                        i++;
                        continue;
                }                       

                if ( i == 0 ) {
                        ptr->procedure = strtoul(plist->str, NULL, 0);
                        ptr->flags |= RPC_CHECK_PROCEDURE;
                }
                
                else if ( i == 1 ) {
                        ptr->version = strtoul(plist->str, NULL, 0);
                        ptr->flags |= RPC_CHECK_VERSION;
                }
                
                else if ( i == 2 ) {
                        ptr->program = strtoul(plist->str, NULL, 0);
                        ptr->flags |= RPC_CHECK_PROGRAM;
                }

                i++;
        }

        if ( ptr->flags )
                return add_rpc_rule(ptr, rules);
        
        free(ptr);
        return -1;
}




static plugin_protocol_t plugin;


static void print_help(void) 
{
        fprintf(stderr, "Usage for %s :\n", plugin.name);
        fprintf(stderr, "\t -p --port-list  List of port this plugin should look at.\n\n");
}



static void setup_own_default(void) 
{
	protocol_plugin_add_port_to_list(&port_list, 111);
}



static int get_default_from_config_file(void) 
{
	int ret = 0;
        config_t *cfg;
        const char *str;
        
        cfg = config_open(PRELUDE_CONF);
        if ( ! cfg ) {
		log(LOG_ERR, "couldn't open %s.\n", PRELUDE_CONF);
		return -1;
	}
        
	str = config_get(cfg, plugin.name, "port_list");
	if ( str ) 
		protocol_plugin_add_string_port_to_list(&port_list, str);
	else 
		ret = -1;

	config_close(cfg);

	return ret;
}




static int check_opts(void) 
{
	char **argv;
        int ret, ok = -1, c, argc;
	struct option opts[] = {
                { "port-list", required_argument, NULL, 'p' },
                { "help", no_argument, NULL, 'h'            },
                { 0, 0, 0, 0 },
        };

        plugin_get_opts(plugin.name, &argc, &argv);
        
        while ( (c = getopt_long(argc, argv, "p:h", opts, NULL) ) != -1 ) {
                switch (c) {

                case 'p':
			ok = 1;
                        ret = protocol_plugin_add_string_port_to_list(&port_list, optarg);
                        if ( ret < 0 )
                                return -1;
                        break;

                case 'h':
                        print_help();
                        break;
                        
                default:
                        return -1;
                }
        }

	if ( ok < 0 && get_default_from_config_file() < 0 )
		setup_own_default();
        
        return 0;
}




int plugin_init(unsigned int id)
{
        int ret;
        
        plugin_set_name(&plugin, "RpcMod");
        protocol_plugin_init_port_list(&port_list);
	
	ret = check_opts();
        if ( ret < 0 )
                return -1;
        
	rpc_plugin_id = id;
        signature_parser_add_multiple_args_key("rpc", &parse_rpc_rule);
        
        plugin_set_protocol(&plugin, "rpc");
        plugin_set_author(&plugin, "Yoann Vandoorselaere");
        plugin_set_contact(&plugin, "yoann@mandrakesoft.com");
        plugin_set_desc(&plugin, "Decode the RPC protocol and provide rpc key for the signature engine.");
        plugin_set_running_func(&plugin, decode_rpc);
        
        return plugin_register((plugin_generic_t *) &plugin);
}
